@@ -8,6 +8,7 @@ package sanitizers_test
 
 import (
 	"bytes"
+	"crypto/fips140"
 	"fmt"
 	"internal/platform"
 	"internal/testenv"
@@ -73,11 +74,11 @@ func TestASAN(t *testing.T) {
 						!strings.Contains(out, noSymbolizer) &&
 						compilerSupportsLocation() {
 
-						t.Errorf("%#q exited without expected location of the error\n%s; got failure\n%s", strings.Join(cmd.Args, " "), tc.errorLocation, out)
+						t.Errorf("%#q exited without expected location of the error\n%s; got failure\n%s", cmd, tc.errorLocation, out)
 					}
 					return
 				}
-				t.Fatalf("%#q exited without expected memory access error\n%s; got failure\n%s", strings.Join(cmd.Args, " "), tc.memoryAccessError, out)
+				t.Fatalf("%#q exited without expected memory access error\n%s; got failure\n%s", cmd, tc.memoryAccessError, out)
 			}
 			mustRun(t, cmd)
 		})
@@ -137,6 +138,9 @@ func TestASANFuzz(t *testing.T) {
 	if bytes.Contains(out, []byte("AddressSanitizer")) {
 		t.Error(`output contains "AddressSanitizer", but should not`)
 	}
+	if !bytes.Contains(out, []byte("FUZZ FAILED")) {
+		t.Error(`fuzz test did not fail with a "FUZZ FAILED" sentinel error`)
+	}
 }
 
 func mustHaveASAN(t *testing.T) *config {
@@ -154,6 +158,10 @@ func mustHaveASAN(t *testing.T) *config {
 		t.Skipf("skipping on %s/%s; -asan option is not supported.", goos, goarch)
 	}
 
+	if fips140.Enabled() {
+		t.Skipf("skipping with FIPS 140 mode; -asan option is not supported.")
+	}
+
 	// The current implementation is only compatible with the ASan library from version
 	// v7 to v9 (See the description in src/runtime/asan/asan.go). Therefore, using the
 	// -asan option must use a compatible version of ASan library, which requires that
