@@ -50,7 +50,7 @@ func (d *deadcodePass) init() {
 		n := d.ldr.NDef()
 		for i := 1; i < n; i++ {
 			s := loader.Sym(i)
-			if d.ldr.SymType(s) == sym.STEXT && d.ldr.SymSize(s) == 0 {
+			if d.ldr.SymType(s).IsText() && d.ldr.SymSize(s) == 0 {
 				// Zero-sized text symbol is a function deadcoded by the
 				// compiler. It doesn't really get compiled, and its
 				// metadata may be missing.
@@ -116,6 +116,13 @@ func (d *deadcodePass) init() {
 		}
 		d.mark(s, 0)
 	}
+	// So are wasmexports.
+	for _, s := range d.ldr.WasmExports {
+		if d.ctxt.Debugvlog > 1 {
+			d.ctxt.Logf("deadcode start wasmexport: %s<%d>\n", d.ldr.SymName(s), d.ldr.SymVersion(s))
+		}
+		d.mark(s, 0)
+	}
 
 	d.mapinitnoop = d.ldr.Lookup("runtime.mapinitnoop", abiInternalVer)
 	if d.mapinitnoop == 0 {
@@ -552,8 +559,15 @@ func (d *deadcodePass) decodetypeMethods(ldr *loader.Loader, arch *sys.Arch, sym
 		off += 3 * arch.PtrSize
 	case abi.Chan: // reflect.chanType
 		off += 2 * arch.PtrSize
-	case abi.Map: // reflect.mapType
-		off += 4*arch.PtrSize + 8
+	case abi.Map:
+		if buildcfg.Experiment.SwissMap {
+			off += 7*arch.PtrSize + 4 // internal/abi.SwissMapType
+			if arch.PtrSize == 8 {
+				off += 4 // padding for final uint32 field (Flags).
+			}
+		} else {
+			off += 4*arch.PtrSize + 8 // internal/abi.OldMapType
+		}
 	case abi.Interface: // reflect.interfaceType
 		off += 3 * arch.PtrSize
 	default:
