@@ -21,13 +21,14 @@ import (
 	"cmd/internal/sys"
 	"cmd/link/internal/loader"
 	"cmd/link/internal/sym"
+	"cmp"
 	"fmt"
 	"internal/abi"
 	"internal/buildcfg"
 	"log"
 	"path"
 	"runtime"
-	"sort"
+	"slices"
 	"strings"
 	"sync"
 )
@@ -147,6 +148,13 @@ func (c dwctxt) AddDWARFAddrSectionOffset(s dwarf.Sym, t interface{}, ofs int64)
 	dsu.AddSymRef(c.arch, tds, ofs, objabi.R_DWARFSECREF, size)
 }
 
+func (c dwctxt) AddIndirectTextRef(s dwarf.Sym, t interface{}) {
+	ds := loader.Sym(s.(dwSym))
+	dsu := c.ldr.MakeSymbolUpdater(ds)
+	tds := loader.Sym(t.(dwSym))
+	dsu.AddSymRef(c.arch, tds, 0, objabi.R_DWTXTADDR_U4, 4)
+}
+
 func (c dwctxt) Logf(format string, args ...interface{}) {
 	c.linkctxt.Logf(format, args...)
 }
@@ -286,7 +294,7 @@ func (d *dwctxt) newdie(parent *dwarf.DWDie, abbrev int, name string) *dwarf.DWD
 
 	var st sym.SymKind
 	switch abbrev {
-	case dwarf.DW_ABRV_FUNCTYPEPARAM, dwarf.DW_ABRV_DOTDOTDOT, dwarf.DW_ABRV_STRUCTFIELD, dwarf.DW_ABRV_ARRAYRANGE:
+	case dwarf.DW_ABRV_FUNCTYPEPARAM, dwarf.DW_ABRV_FUNCTYPEOUTPARAM, dwarf.DW_ABRV_DOTDOTDOT, dwarf.DW_ABRV_STRUCTFIELD, dwarf.DW_ABRV_ARRAYRANGE:
 		// There are no relocations against these dies, and their names
 		// are not unique, so don't create a symbol.
 		return die
@@ -479,7 +487,7 @@ func (d *dwctxt) dotypedef(parent *dwarf.DWDie, name string, def *dwarf.DWDie) *
 		return nil
 	}
 	if def == nil {
-		Errorf(nil, "dwarf: bad def in dotypedef")
+		Errorf("dwarf: bad def in dotypedef")
 	}
 
 	// Create a new loader symbol for the typedef. We no longer
@@ -519,7 +527,7 @@ func (d *dwctxt) defgotype(gotype loader.Sym) loader.Sym {
 		d.linkctxt.Errorf(gotype, "dwarf: type name doesn't start with \"type:\"")
 		return d.mustFind("<unspecified>")
 	}
-	name := sn[5:] // could also decode from Type.string
+	name := sn[len("type:"):] // could also decode from Type.string
 
 	sdie := d.find(name)
 	if sdie != 0 {
@@ -533,7 +541,7 @@ func (d *dwctxt) defgotype(gotype loader.Sym) loader.Sym {
 
 func (d *dwctxt) newtype(gotype loader.Sym) *dwarf.DWDie {
 	sn := d.ldr.SymName(gotype)
-	name := sn[5:] // could also decode from Type.string
+	name := sn[len("type:"):] // could also decode from Type.string
 	tdata := d.ldr.Data(gotype)
 	if len(tdata) == 0 {
 		d.linkctxt.Errorf(gotype, "missing type")
@@ -583,7 +591,7 @@ func (d *dwctxt) newtype(gotype loader.Sym) *dwarf.DWDie {
 		die = d.newdie(&dwtypes, dwarf.DW_ABRV_ARRAYTYPE, name)
 		typedefdie = d.dotypedef(&dwtypes, name, die)
 		newattr(die, dwarf.DW_AT_byte_size, dwarf.DW_CLS_CONSTANT, bytesize, 0)
-		s := decodetypeArrayElem(d.ldr, d.arch, gotype)
+		s := decodetypeArrayElem(d.linkctxt, d.arch, gotype)
 		d.newrefattr(die, dwarf.DW_AT_type, d.defgotype(s))
 		fld := d.newdie(die, dwarf.DW_ABRV_ARRAYRANGE, "range")
 
@@ -622,8 +630,9 @@ func (d *dwctxt) newtype(gotype loader.Sym) *dwarf.DWDie {
 		for i := 0; i < nfields; i++ {
 			s := decodetypeFuncOutType(d.ldr, d.arch, gotype, &relocs, i)
 			sn := d.ldr.SymName(s)
-			fld := d.newdie(die, dwarf.DW_ABRV_FUNCTYPEPARAM, sn[5:])
-			d.newrefattr(fld, dwarf.DW_AT_type, d.defptrto(d.defgotype(s)))
+			fld := d.newdie(die, dwarf.DW_ABRV_FUNCTYPEOUTPARAM, sn[5:])
+			newattr(fld, dwarf.DW_AT_variable_parameter, dwarf.DW_CLS_FLAG, 1, 0)
+			d.newrefattr(fld, dwarf.DW_AT_type, d.defgotype(s))
 		}
 
 	case abi.Interface:
@@ -659,7 +668,7 @@ func (d *dwctxt) newtype(gotype loader.Sym) *dwarf.DWDie {
 		die = d.newdie(&dwtypes, dwarf.DW_ABRV_SLICETYPE, name)
 		typedefdie = d.dotypedef(&dwtypes, name, die)
 		newattr(die, dwarf.DW_AT_byte_size, dwarf.DW_CLS_CONSTANT, bytesize, 0)
-		s := decodetypeArrayElem(d.ldr, d.arch, gotype)
+		s := decodetypeArrayElem(d.linkctxt, d.arch, gotype)
 		elem := d.defgotype(s)
 		d.newrefattr(die, dwarf.DW_AT_go_elem, elem)
 
@@ -674,7 +683,7 @@ func (d *dwctxt) newtype(gotype loader.Sym) *dwarf.DWDie {
 		nfields := decodetypeStructFieldCount(d.ldr, d.arch, gotype)
 		for i := 0; i < nfields; i++ {
 			f := decodetypeStructFieldName(d.ldr, d.arch, gotype, i)
-			s := decodetypeStructFieldType(d.ldr, d.arch, gotype, i)
+			s := decodetypeStructFieldType(d.linkctxt, d.arch, gotype, i)
 			if f == "" {
 				sn := d.ldr.SymName(s)
 				f = sn[5:] // skip "type:"
@@ -863,6 +872,76 @@ func (d *dwctxt) mkinternaltype(ctxt *Link, abbrev int, typename, keyname, valna
 }
 
 func (d *dwctxt) synthesizemaptypes(ctxt *Link, die *dwarf.DWDie) {
+	if buildcfg.Experiment.SwissMap {
+		d.synthesizemaptypesSwiss(ctxt, die)
+	} else {
+		d.synthesizemaptypesOld(ctxt, die)
+	}
+}
+
+func (d *dwctxt) synthesizemaptypesSwiss(ctxt *Link, die *dwarf.DWDie) {
+	mapType := walktypedef(d.findprotodie(ctxt, "type:internal/runtime/maps.Map"))
+	tableType := walktypedef(d.findprotodie(ctxt, "type:internal/runtime/maps.table"))
+	groupsReferenceType := walktypedef(d.findprotodie(ctxt, "type:internal/runtime/maps.groupsReference"))
+
+	for ; die != nil; die = die.Link {
+		if die.Abbrev != dwarf.DW_ABRV_MAPTYPE {
+			continue
+		}
+		gotype := loader.Sym(getattr(die, dwarf.DW_AT_type).Data.(dwSym))
+
+		keyType := decodetypeMapKey(d.ldr, d.arch, gotype)
+		valType := decodetypeMapValue(d.ldr, d.arch, gotype)
+		groupType := decodetypeMapSwissGroup(d.ldr, d.arch, gotype)
+
+		keyType = d.walksymtypedef(d.defgotype(keyType))
+		valType = d.walksymtypedef(d.defgotype(valType))
+		groupType = d.walksymtypedef(d.defgotype(groupType))
+
+		keyName := d.nameFromDIESym(keyType)
+		valName := d.nameFromDIESym(valType)
+
+		// Construct groupsReference[K,V]
+		dwGroupsReference := d.mkinternaltype(ctxt, dwarf.DW_ABRV_STRUCTTYPE, "groupReference", keyName, valName, func(dwh *dwarf.DWDie) {
+			d.copychildren(ctxt, dwh, groupsReferenceType)
+			// data *group[K,V]
+			//
+			// This is actually a pointer to an array
+			// *[lengthMask+1]group[K,V], but the length is
+			// variable, so we can't statically record the length.
+			d.substitutetype(dwh, "data", d.defptrto(groupType))
+			newattr(dwh, dwarf.DW_AT_byte_size, dwarf.DW_CLS_CONSTANT, getattr(groupsReferenceType, dwarf.DW_AT_byte_size).Value, nil)
+			newattr(dwh, dwarf.DW_AT_go_kind, dwarf.DW_CLS_CONSTANT, int64(abi.Struct), 0)
+		})
+
+		// Construct table[K,V]
+		dwTable := d.mkinternaltype(ctxt, dwarf.DW_ABRV_STRUCTTYPE, "table", keyName, valName, func(dwh *dwarf.DWDie) {
+			d.copychildren(ctxt, dwh, tableType)
+			d.substitutetype(dwh, "groups", dwGroupsReference)
+			newattr(dwh, dwarf.DW_AT_byte_size, dwarf.DW_CLS_CONSTANT, getattr(tableType, dwarf.DW_AT_byte_size).Value, nil)
+			newattr(dwh, dwarf.DW_AT_go_kind, dwarf.DW_CLS_CONSTANT, int64(abi.Struct), 0)
+		})
+
+		// Construct map[K,V]
+		dwMap := d.mkinternaltype(ctxt, dwarf.DW_ABRV_STRUCTTYPE, "map", keyName, valName, func(dwh *dwarf.DWDie) {
+			d.copychildren(ctxt, dwh, mapType)
+			// dirPtr is a pointer to a variable-length array of
+			// *table[K,V], of length dirLen.
+			//
+			// Since we can't directly define a variable-length
+			// array, store this as **table[K,V]. i.e., pointer to
+			// the first entry in the array.
+			d.substitutetype(dwh, "dirPtr", d.defptrto(d.defptrto(dwTable)))
+			newattr(dwh, dwarf.DW_AT_byte_size, dwarf.DW_CLS_CONSTANT, getattr(mapType, dwarf.DW_AT_byte_size).Value, nil)
+			newattr(dwh, dwarf.DW_AT_go_kind, dwarf.DW_CLS_CONSTANT, int64(abi.Struct), 0)
+		})
+
+		// make map type a pointer to map[K,V]
+		d.newrefattr(die, dwarf.DW_AT_type, d.defptrto(dwMap))
+	}
+}
+
+func (d *dwctxt) synthesizemaptypesOld(ctxt *Link, die *dwarf.DWDie) {
 	hash := walktypedef(d.findprotodie(ctxt, "type:runtime.hmap"))
 	bucket := walktypedef(d.findprotodie(ctxt, "type:runtime.bmap"))
 
@@ -884,11 +963,11 @@ func (d *dwctxt) synthesizemaptypes(ctxt *Link, die *dwarf.DWDie) {
 
 		// compute size info like hashmap.c does.
 		indirectKey, indirectVal := false, false
-		if keysize > abi.MapMaxKeyBytes {
+		if keysize > abi.OldMapMaxKeyBytes {
 			keysize = int64(d.arch.PtrSize)
 			indirectKey = true
 		}
-		if valsize > abi.MapMaxElemBytes {
+		if valsize > abi.OldMapMaxElemBytes {
 			valsize = int64(d.arch.PtrSize)
 			indirectVal = true
 		}
@@ -896,28 +975,28 @@ func (d *dwctxt) synthesizemaptypes(ctxt *Link, die *dwarf.DWDie) {
 		// Construct type to represent an array of BucketSize keys
 		keyname := d.nameFromDIESym(keytype)
 		dwhks := d.mkinternaltype(ctxt, dwarf.DW_ABRV_ARRAYTYPE, "[]key", keyname, "", func(dwhk *dwarf.DWDie) {
-			newattr(dwhk, dwarf.DW_AT_byte_size, dwarf.DW_CLS_CONSTANT, abi.MapBucketCount*keysize, 0)
+			newattr(dwhk, dwarf.DW_AT_byte_size, dwarf.DW_CLS_CONSTANT, abi.OldMapBucketCount*keysize, 0)
 			t := keytype
 			if indirectKey {
 				t = d.defptrto(keytype)
 			}
 			d.newrefattr(dwhk, dwarf.DW_AT_type, t)
 			fld := d.newdie(dwhk, dwarf.DW_ABRV_ARRAYRANGE, "size")
-			newattr(fld, dwarf.DW_AT_count, dwarf.DW_CLS_CONSTANT, abi.MapBucketCount, 0)
+			newattr(fld, dwarf.DW_AT_count, dwarf.DW_CLS_CONSTANT, abi.OldMapBucketCount, 0)
 			d.newrefattr(fld, dwarf.DW_AT_type, d.uintptrInfoSym)
 		})
 
 		// Construct type to represent an array of BucketSize values
 		valname := d.nameFromDIESym(valtype)
 		dwhvs := d.mkinternaltype(ctxt, dwarf.DW_ABRV_ARRAYTYPE, "[]val", valname, "", func(dwhv *dwarf.DWDie) {
-			newattr(dwhv, dwarf.DW_AT_byte_size, dwarf.DW_CLS_CONSTANT, abi.MapBucketCount*valsize, 0)
+			newattr(dwhv, dwarf.DW_AT_byte_size, dwarf.DW_CLS_CONSTANT, abi.OldMapBucketCount*valsize, 0)
 			t := valtype
 			if indirectVal {
 				t = d.defptrto(valtype)
 			}
 			d.newrefattr(dwhv, dwarf.DW_AT_type, t)
 			fld := d.newdie(dwhv, dwarf.DW_ABRV_ARRAYRANGE, "size")
-			newattr(fld, dwarf.DW_AT_count, dwarf.DW_CLS_CONSTANT, abi.MapBucketCount, 0)
+			newattr(fld, dwarf.DW_AT_count, dwarf.DW_CLS_CONSTANT, abi.OldMapBucketCount, 0)
 			d.newrefattr(fld, dwarf.DW_AT_type, d.uintptrInfoSym)
 		})
 
@@ -929,20 +1008,20 @@ func (d *dwctxt) synthesizemaptypes(ctxt *Link, die *dwarf.DWDie) {
 
 			fld := d.newdie(dwhb, dwarf.DW_ABRV_STRUCTFIELD, "keys")
 			d.newrefattr(fld, dwarf.DW_AT_type, dwhks)
-			newmemberoffsetattr(fld, abi.MapBucketCount)
+			newmemberoffsetattr(fld, abi.OldMapBucketCount)
 			fld = d.newdie(dwhb, dwarf.DW_ABRV_STRUCTFIELD, "values")
 			d.newrefattr(fld, dwarf.DW_AT_type, dwhvs)
-			newmemberoffsetattr(fld, abi.MapBucketCount+abi.MapBucketCount*int32(keysize))
+			newmemberoffsetattr(fld, abi.OldMapBucketCount+abi.OldMapBucketCount*int32(keysize))
 			fld = d.newdie(dwhb, dwarf.DW_ABRV_STRUCTFIELD, "overflow")
 			d.newrefattr(fld, dwarf.DW_AT_type, d.defptrto(d.dtolsym(dwhb.Sym)))
-			newmemberoffsetattr(fld, abi.MapBucketCount+abi.MapBucketCount*(int32(keysize)+int32(valsize)))
+			newmemberoffsetattr(fld, abi.OldMapBucketCount+abi.OldMapBucketCount*(int32(keysize)+int32(valsize)))
 			if d.arch.RegSize > d.arch.PtrSize {
 				fld = d.newdie(dwhb, dwarf.DW_ABRV_STRUCTFIELD, "pad")
 				d.newrefattr(fld, dwarf.DW_AT_type, d.uintptrInfoSym)
-				newmemberoffsetattr(fld, abi.MapBucketCount+abi.MapBucketCount*(int32(keysize)+int32(valsize))+int32(d.arch.PtrSize))
+				newmemberoffsetattr(fld, abi.OldMapBucketCount+abi.OldMapBucketCount*(int32(keysize)+int32(valsize))+int32(d.arch.PtrSize))
 			}
 
-			newattr(dwhb, dwarf.DW_AT_byte_size, dwarf.DW_CLS_CONSTANT, abi.MapBucketCount+abi.MapBucketCount*keysize+abi.MapBucketCount*valsize+int64(d.arch.RegSize), 0)
+			newattr(dwhb, dwarf.DW_AT_byte_size, dwarf.DW_CLS_CONSTANT, abi.OldMapBucketCount+abi.OldMapBucketCount*keysize+abi.OldMapBucketCount*valsize+int64(d.arch.RegSize), 0)
 		})
 
 		// Construct hash<K,V>
@@ -1140,18 +1219,70 @@ func expandFile(fname string) string {
 
 // writeDirFileTables emits the portion of the DWARF line table
 // prologue containing the include directories and file names,
-// described in section 6.2.4 of the DWARF 4 standard. It walks the
+// described in section 6.2.4 of the DWARF standard. It walks the
 // filepaths for the unit to discover any common directories, which
 // are emitted to the directory table first, then the file table is
 // emitted after that.
+//
+// Note that there are some differences betwen DWARF versions 4 and 5
+// regarding how files and directories are handled; the chief item is
+// that in version 4 we have an implicit directory index 0 that holds
+// the compilation dir, and an implicit file index 0 that holds the
+// "primary" source file being compiled (with the assumption that
+// we'll get these two pieces of info from attributes on the
+// compilation unit DIE). DWARF version 5 does away with these
+// implicit entries; if you want to refer to a file or dir, you have
+// to mention it explicitly.
+//
+// So, let's say we have a Go package with import path
+// "github.com/fruit/bowl" with two source files, "apple.go" and
+// "orange.go". The directory and file table in version 4 might look
+// like:
+//
+//	Dir Table:
+//	1		github.com/fruit/bowl
+//
+//	File Table:
+//	Entry	Dir	Time	Size	Name
+//	1		0	0		0		<autogenerated>
+//	2		1	0		0		apple.go
+//	3		1	0		0		orange.go
+//
+// With DWARF version 5, the file and directory tables might look like
+//
+//	Dir Table:
+//	0		.
+//	1		github.com/fruit/bowl
+//
+//	File Table:
+//	Entry	Dir	  Name
+//	0		0	  ?                /* see remark below about this entry */
+//	1		0	  <autogenerated>
+//	2		1	  apple.go
+//	3		1	  orange.go
+//
+// Also worth noting that the line table prolog allows you to control
+// what items or fields appear in file and die entries (as opposed
+// to the fixed dir/time/size/name format for files in DWARF 4).
 func (d *dwctxt) writeDirFileTables(unit *sym.CompilationUnit, lsu *loader.SymbolBuilder) {
 	type fileDir struct {
 		base string
 		dir  int
 	}
 	dirNums := make(map[string]int)
-	dirs := []string{""}
+	dirs := []string{"."}
+	dirNums["."] = 0
 	files := []fileDir{}
+	if buildcfg.Experiment.Dwarf5 {
+		// Add a dummy zero entry (not used for anything) so as to
+		// ensure that the first useful file index is 1, since that's
+		// the default value of the line table file register).  Note
+		// the "?"; this was originally an empty string, but doing
+		// this triggers crashes in some versions of GDB, see
+		// https://sourceware.org/bugzilla/show_bug.cgi?id=30357 for
+		// the details.
+		files = append(files, fileDir{base: "?", dir: 0})
+	}
 
 	// Preprocess files to collect directories. This assumes that the
 	// file table is already de-duped.
@@ -1169,7 +1300,7 @@ func (d *dwctxt) writeDirFileTables(unit *sym.CompilationUnit, lsu *loader.Symbo
 		dir := path.Dir(name)
 		dirIdx, ok := dirNums[dir]
 		if !ok && dir != "." {
-			dirIdx = len(dirNums) + 1
+			dirIdx = len(dirNums)
 			dirNums[dir] = dirIdx
 			dirs = append(dirs, dir)
 		}
@@ -1188,22 +1319,61 @@ func (d *dwctxt) writeDirFileTables(unit *sym.CompilationUnit, lsu *loader.Symbo
 		}
 	}
 
-	// Emit directory section. This is a series of nul terminated
-	// strings, followed by a single zero byte.
 	lsDwsym := dwSym(lsu.Sym())
-	for k := 1; k < len(dirs); k++ {
-		d.AddString(lsDwsym, dirs[k])
-	}
-	lsu.AddUint8(0) // terminator
+	if buildcfg.Experiment.Dwarf5 {
+		// Emit DWARF5 directory and file sections. The key added
+		// element here for version 5 is that we emit a small prolog
+		// describing the format of each file/dir entry, then a count
+		// of the entries, then the entries themselves.
 
-	// Emit file section.
-	for k := 0; k < len(files); k++ {
-		d.AddString(lsDwsym, files[k].base)
-		dwarf.Uleb128put(d, lsDwsym, int64(files[k].dir))
-		lsu.AddUint8(0) // mtime
-		lsu.AddUint8(0) // length
+		// Dir table first...
+		lsu.AddUint8(1) // directory_entry_format_count
+		// each directory entry is just a single path string.
+		dwarf.Uleb128put(d, lsDwsym, dwarf.DW_LNCT_path)
+		dwarf.Uleb128put(d, lsDwsym, dwarf.DW_FORM_string)
+		// emit count, then dirs
+		dwarf.Uleb128put(d, lsDwsym, int64(len(dirs)))
+		for k := 0; k < len(dirs); k++ {
+			d.AddString(lsDwsym, dirs[k])
+		}
+
+		// ... now file table.  With DWARF version 5 we put out a
+		// prolog that describes the format of each file entry; our
+		// chosen format is a pair <F,D> where F is the file and D is
+		// the dir index (zero-based).
+		lsu.AddUint8(2) // file_entry_format_count
+		// each file entry is just a file path followed by dir index.
+		dwarf.Uleb128put(d, lsDwsym, dwarf.DW_LNCT_path)
+		dwarf.Uleb128put(d, lsDwsym, dwarf.DW_FORM_string)
+		dwarf.Uleb128put(d, lsDwsym, dwarf.DW_LNCT_directory_index)
+		dwarf.Uleb128put(d, lsDwsym, dwarf.DW_FORM_udata)
+		// emit count, then files
+		dwarf.Uleb128put(d, lsDwsym, int64(len(files)))
+		for k := 0; k < len(files); k++ {
+			d.AddString(lsDwsym, files[k].base)
+			dwarf.Uleb128put(d, lsDwsym, int64(files[k].dir))
+		}
+	} else {
+		// Emit DWARF4 directory and file sections.
+
+		// Dir table first. This just a series of nul terminated
+		// strings, followed by a single zero byte.
+		for k := 1; k < len(dirs); k++ {
+			d.AddString(lsDwsym, dirs[k])
+		}
+		lsu.AddUint8(0) // terminator needed in V4 case.
+
+		// File table next. This is a series of
+		// file/dirindex/mtime/length tuples (where dirindex is
+		// 1-based) followed by a single zero byte.
+		for k := 0; k < len(files); k++ {
+			d.AddString(lsDwsym, files[k].base)
+			dwarf.Uleb128put(d, lsDwsym, int64(files[k].dir))
+			lsu.AddUint8(0) // mtime
+			lsu.AddUint8(0) // length
+		}
+		lsu.AddUint8(0) // terminator
 	}
-	lsu.AddUint8(0) // terminator
 }
 
 // writelines collects up and chains together the symbols needed to
@@ -1232,13 +1402,23 @@ func (d *dwctxt) writelines(unit *sym.CompilationUnit, lineProlog loader.Sym) []
 	unitLengthOffset := lsu.Size()
 	d.createUnitLength(lsu, 0) // unit_length (*), filled in at end
 	unitstart = lsu.Size()
-	lsu.AddUint16(d.arch, 2) // dwarf version (appendix F) -- version 3 is incompatible w/ XCode 9.0's dsymutil, latest supported on OSX 10.12 as of 2018-05
+	if buildcfg.Experiment.Dwarf5 {
+		lsu.AddUint16(d.arch, 5)
+		// DWARF5 requires address_size and segment selector sizes
+		// here, both ubyte format.
+		lsu.AddUint8(uint8(d.arch.PtrSize))
+		lsu.AddUint8(0)
+	} else {
+		lsu.AddUint16(d.arch, 2) // dwarf version (appendix F) -- version 3 is incompatible w/ XCode 9.0's dsymutil, latest supported on OSX 10.12 as of 2018-05
+	}
 	headerLengthOffset := lsu.Size()
 	d.addDwarfAddrField(lsu, 0) // header_length (*), filled in at end
 	headerstart = lsu.Size()
 
-	// cpos == unitstart + 4 + 2 + 4
-	lsu.AddUint8(1)                // minimum_instruction_length
+	lsu.AddUint8(1) // minimum_instruction_length
+	if buildcfg.Experiment.Dwarf5 {
+		lsu.AddUint8(1) // maximum_operations_per_instruction
+	}
 	lsu.AddUint8(is_stmt)          // default_is_stmt
 	lsu.AddUint8(LINE_BASE & 0xFF) // line_base
 	lsu.AddUint8(LINE_RANGE)       // line_range
@@ -1291,7 +1471,7 @@ func (d *dwctxt) writelines(unit *sym.CompilationUnit, lineProlog loader.Sym) []
 // writepcranges generates the DW_AT_ranges table for compilation unit
 // "unit", and returns a collection of ranges symbols (one for the
 // compilation unit DIE itself and the remainder from functions in the unit).
-func (d *dwctxt) writepcranges(unit *sym.CompilationUnit, base loader.Sym, pcs []dwarf.Range, rangeProlog loader.Sym) []loader.Sym {
+func (d *dwctxt) writepcranges(unit *sym.CompilationUnit, base loader.Sym, pcs []dwarf.Range, rangeProlog loader.Sym, debugaddrsym loader.Sym) []loader.Sym {
 
 	syms := make([]loader.Sym, 0, len(unit.RangeSyms)+1)
 	syms = append(syms, rangeProlog)
@@ -1301,7 +1481,24 @@ func (d *dwctxt) writepcranges(unit *sym.CompilationUnit, base loader.Sym, pcs [
 	// Create PC ranges for the compilation unit DIE.
 	newattr(unit.DWInfo, dwarf.DW_AT_ranges, dwarf.DW_CLS_PTR, rsu.Size(), rDwSym)
 	newattr(unit.DWInfo, dwarf.DW_AT_low_pc, dwarf.DW_CLS_ADDRESS, 0, dwSym(base))
-	dwarf.PutBasedRanges(d, rDwSym, pcs)
+	if buildcfg.Experiment.Dwarf5 && debugaddrsym != 0 {
+		debugaddr := d.ldr.MakeSymbolUpdater(debugaddrsym)
+		// Write DWARF5-based ranges for the unit. This will introduce
+		// a series of new R_DWTXTADDR_* relocs, which we'll process
+		// in the loop immediately following this call.
+		dwarf.PutRngListRanges(d, rDwSym, dwSym(base), pcs)
+		drelocs := d.ldr.Relocs(rangeProlog)
+		for ri := 0; ri < drelocs.Count(); ri++ {
+			r := drelocs.At(ri)
+			if !r.Type().IsDwTxtAddr() {
+				continue
+			}
+			cusym := d.dtolsym(unit.DWInfo.Sym)
+			d.assignDebugAddrSlot(unit, cusym, r, debugaddr)
+		}
+	} else {
+		dwarf.PutBasedRanges(d, rDwSym, pcs)
+	}
 
 	// Collect up the ranges for functions in the unit.
 	rsize := uint64(rsu.Size())
@@ -1500,11 +1697,7 @@ func (d *dwctxt) writeframes(fs loader.Sym) dwarfSecInfo {
  *  Walk DWarfDebugInfoEntries, and emit .debug_info
  */
 
-const (
-	COMPUNITHEADERSIZE = 4 + 2 + 4 + 1
-)
-
-func (d *dwctxt) writeUnitInfo(u *sym.CompilationUnit, abbrevsym loader.Sym, infoEpilog loader.Sym) []loader.Sym {
+func (d *dwctxt) writeUnitInfo(u *sym.CompilationUnit, abbrevsym loader.Sym, addrsym loader.Sym, infoEpilog loader.Sym) []loader.Sym {
 	syms := []loader.Sym{}
 	if len(u.Textp) == 0 && u.DWInfo.Child == nil && len(u.VarDIEs) == 0 {
 		return syms
@@ -1516,17 +1709,39 @@ func (d *dwctxt) writeUnitInfo(u *sym.CompilationUnit, abbrevsym loader.Sym, inf
 
 	// Write .debug_info Compilation Unit Header (sec 7.5.1)
 	// Fields marked with (*) must be changed for 64-bit dwarf
-	// This must match COMPUNITHEADERSIZE above.
 	d.createUnitLength(su, 0) // unit_length (*), will be filled in later.
-	su.AddUint16(d.arch, 4)   // dwarf version (appendix F)
+	dwver := 4
+	if buildcfg.Experiment.Dwarf5 {
+		dwver = 5
+	}
+	su.AddUint16(d.arch, uint16(dwver)) // dwarf version (appendix F)
 
-	// debug_abbrev_offset (*)
-	d.addDwarfAddrRef(su, abbrevsym)
-
-	su.AddUint8(uint8(d.arch.PtrSize)) // address_size
+	if buildcfg.Experiment.Dwarf5 {
+		// DWARF5 at this point requires
+		// 1. unit type
+		// 2. address size
+		// 3. abbrev offset
+		su.AddUint8(uint8(dwarf.DW_UT_compile))
+		su.AddUint8(uint8(d.arch.PtrSize))
+		d.addDwarfAddrRef(su, abbrevsym)
+	} else {
+		// DWARF4 requires
+		// 1. abbrev offset
+		// 2. address size
+		d.addDwarfAddrRef(su, abbrevsym)
+		su.AddUint8(uint8(d.arch.PtrSize))
+	}
 
 	ds := dwSym(s)
 	dwarf.Uleb128put(d, ds, int64(compunit.Abbrev))
+	if buildcfg.Experiment.Dwarf5 {
+		// If this CU has functions, update the DW_AT_addr_base
+		// attribute to point to the correct section symbol.
+		abattr := getattr(compunit, dwarf.DW_AT_addr_base)
+		if abattr != nil {
+			abattr.Data = dwSym(addrsym)
+		}
+	}
 	dwarf.PutAttrs(d, ds, compunit.Abbrev, compunit.Attr)
 
 	// This is an under-estimate; more will be needed for type DIEs.
@@ -1652,6 +1867,35 @@ func (d *dwctxt) mkBuiltinType(ctxt *Link, abrv int, tname string) *dwarf.DWDie
 	return die
 }
 
+// assignDebugAddrSlot assigns a slot (if needed) in the .debug_addr
+// section fragment of the specified unit for the function pointed to
+// by R_DWTXTADDR_* relocation r.  The slot index selected will be
+// filled in when the relocation is actually applied/resolved.
+func (d *dwctxt) assignDebugAddrSlot(unit *sym.CompilationUnit, fnsym loader.Sym, r loader.Reloc, sb *loader.SymbolBuilder) {
+	rsym := r.Sym()
+	if unit.Addrs == nil {
+		unit.Addrs = make(map[sym.LoaderSym]uint32)
+	}
+	if _, ok := unit.Addrs[sym.LoaderSym(rsym)]; ok {
+		// already present, no work needed
+	} else {
+		sl := len(unit.Addrs)
+		rt := r.Type()
+		lim, _ := rt.DwTxtAddrRelocParams()
+		if sl > lim {
+			log.Fatalf("internal error: %s relocation overflow on infosym for %s", rt.String(), d.ldr.SymName(fnsym))
+		}
+		unit.Addrs[sym.LoaderSym(rsym)] = uint32(sl)
+		sb.AddAddrPlus(d.arch, rsym, 0)
+		data := sb.Data()
+		if d.arch.PtrSize == 4 {
+			d.arch.ByteOrder.PutUint32(data[len(data)-4:], uint32(d.ldr.SymValue(rsym)))
+		} else {
+			d.arch.ByteOrder.PutUint64(data[len(data)-8:], uint64(d.ldr.SymValue(rsym)))
+		}
+	}
+}
+
 // dwarfVisitFunction takes a function (text) symbol and processes the
 // subprogram DIE for the function and picks up any other DIEs
 // (absfns, types) that it references.
@@ -1768,12 +2012,18 @@ func dwarfGenerateDebugInfo(ctxt *Link) {
 	prototypedies = map[string]*dwarf.DWDie{
 		"type:runtime.stringStructDWARF": nil,
 		"type:runtime.slice":             nil,
-		"type:runtime.hmap":              nil,
-		"type:runtime.bmap":              nil,
 		"type:runtime.sudog":             nil,
 		"type:runtime.waitq":             nil,
 		"type:runtime.hchan":             nil,
 	}
+	if buildcfg.Experiment.SwissMap {
+		prototypedies["type:internal/runtime/maps.Map"] = nil
+		prototypedies["type:internal/runtime/maps.table"] = nil
+		prototypedies["type:internal/runtime/maps.groupsReference"] = nil
+	} else {
+		prototypedies["type:runtime.hmap"] = nil
+		prototypedies["type:runtime.bmap"] = nil
+	}
 
 	// Needed by the prettyprinter code for interface inspection.
 	for _, typ := range []string{
@@ -1781,7 +2031,6 @@ func dwarfGenerateDebugInfo(ctxt *Link) {
 		"type:internal/abi.ArrayType",
 		"type:internal/abi.ChanType",
 		"type:internal/abi.FuncType",
-		"type:internal/abi.MapType",
 		"type:internal/abi.PtrType",
 		"type:internal/abi.SliceType",
 		"type:internal/abi.StructType",
@@ -1790,6 +2039,11 @@ func dwarfGenerateDebugInfo(ctxt *Link) {
 		"type:internal/abi.Imethod"} {
 		d.defgotype(d.lookupOrDiag(typ))
 	}
+	if buildcfg.Experiment.SwissMap {
+		d.defgotype(d.lookupOrDiag("type:internal/abi.SwissMapType"))
+	} else {
+		d.defgotype(d.lookupOrDiag("type:internal/abi.OldMapType"))
+	}
 
 	// fake root DIE for compile unit DIEs
 	var dwroot dwarf.DWDie
@@ -1852,6 +2106,16 @@ func dwarfGenerateDebugInfo(ctxt *Link) {
 			}
 			newattr(unit.DWInfo, dwarf.DW_AT_go_package_name, dwarf.DW_CLS_STRING, int64(len(pkgname)), pkgname)
 
+			if buildcfg.Experiment.Dwarf5 && cuabrv == dwarf.DW_ABRV_COMPUNIT {
+				// For DWARF5, the CU die will have an attribute that
+				// points to the offset into the .debug_addr section
+				// that contains the compilation unit's
+				// contributions. Add this attribute now. Note that
+				// we'll later on update the Data field in this attr
+				// once we know the .debug_addr sym for the unit.
+				newattr(unit.DWInfo, dwarf.DW_AT_addr_base, dwarf.DW_CLS_REFERENCE, 0, 0)
+			}
+
 			// Scan all functions in this compilation unit, create
 			// DIEs for all referenced types, find all referenced
 			// abstract functions, visit range symbols. Note that
@@ -1883,8 +2147,9 @@ func dwarfGenerateDebugInfo(ctxt *Link) {
 			continue
 		}
 		t := d.ldr.SymType(idx)
-		switch t {
-		case sym.SRODATA, sym.SDATA, sym.SNOPTRDATA, sym.STYPE, sym.SBSS, sym.SNOPTRBSS, sym.STLSBSS:
+		switch {
+		case t.IsRODATA(), t.IsDATA(), t.IsNOPTRDATA(),
+			t == sym.STYPE, t == sym.SBSS, t == sym.SNOPTRBSS, t == sym.STLSBSS:
 			// ok
 		default:
 			continue
@@ -1933,7 +2198,7 @@ func dwarfGenerateDebugInfo(ctxt *Link) {
 
 // dwarfGenerateDebugSyms constructs debug_line, debug_frame, and
 // debug_loc. It also writes out the debug_info section using symbols
-// generated in dwarfGenerateDebugInfo2.
+// generated in dwarfGenerateDebugInfo.
 func dwarfGenerateDebugSyms(ctxt *Link) {
 	if !dwarfEnabled(ctxt) {
 		return
@@ -1960,28 +2225,85 @@ type dwUnitSyms struct {
 	infosyms   []loader.Sym
 	locsyms    []loader.Sym
 	rangessyms []loader.Sym
+	addrsym    loader.Sym
 }
 
-// dwUnitPortion assembles the DWARF content for a given compilation
-// unit: debug_info, debug_lines, debug_ranges, debug_loc (debug_frame
-// is handled elsewhere). Order is important; the calls to writelines
-// and writepcranges below make updates to the compilation unit DIE,
-// hence they have to happen before the call to writeUnitInfo.
+// dwUnitPortion assembles the DWARF content for a given comp unit:
+// debug_info, debug_lines, debug_ranges(V4) or debug_rnglists (V5),
+// debug_loc (V4) or debug_loclists (V5) and debug_addr (V5);
+// debug_frame is handled elsewhere. Order is important; the calls to
+// writelines and writepcranges below make updates to the compilation
+// unit DIE, hence they have to happen before the call to
+// writeUnitInfo.
 func (d *dwctxt) dwUnitPortion(u *sym.CompilationUnit, abbrevsym loader.Sym, us *dwUnitSyms) {
 	if u.DWInfo.Abbrev != dwarf.DW_ABRV_COMPUNIT_TEXTLESS {
 		us.linesyms = d.writelines(u, us.lineProlog)
 		base := loader.Sym(u.Textp[0])
-		us.rangessyms = d.writepcranges(u, base, u.PCs, us.rangeProlog)
+		if buildcfg.Experiment.Dwarf5 {
+			d.writedebugaddr(u, us.addrsym)
+		}
+		us.rangessyms = d.writepcranges(u, base, u.PCs, us.rangeProlog, us.addrsym)
 		us.locsyms = d.collectUnitLocs(u)
 	}
-	us.infosyms = d.writeUnitInfo(u, abbrevsym, us.infoEpilog)
+	us.infosyms = d.writeUnitInfo(u, abbrevsym, us.addrsym, us.infoEpilog)
+}
+
+// writedebugaddr scans the symbols of interest in unit for
+// R_DWTXTADDR_R* relocations and converts these into the material
+// we'll need to generate the .debug_addr section for the unit. This
+// will create a map within the unit (as a side effect), mapping func
+// symbol to debug_addr slot.
+func (d *dwctxt) writedebugaddr(unit *sym.CompilationUnit, debugaddr loader.Sym) {
+	dasu := d.ldr.MakeSymbolUpdater(debugaddr)
+
+	var dsyms []loader.Sym
+	for _, s := range unit.Textp {
+		fnSym := loader.Sym(s)
+		// NB: this looks at SDWARFFCN; it will need to also look
+		// at range and loc when they get there.
+		infosym, locsym, rangessym, _ := d.ldr.GetFuncDwarfAuxSyms(fnSym)
+
+		// Walk the relocations of the various DWARF symbols to
+		// collect relocations corresponding to indirect function
+		// references via .debug_addr.
+		dsyms = dsyms[:0]
+		dsyms = append(dsyms, infosym)
+		if rangessym != 0 {
+			dsyms = append(dsyms, rangessym)
+		}
+		if locsym != 0 {
+			dsyms = append(dsyms, locsym)
+		}
+		for _, dsym := range dsyms {
+			drelocs := d.ldr.Relocs(dsym)
+			for ri := 0; ri < drelocs.Count(); ri++ {
+				r := drelocs.At(ri)
+				if !r.Type().IsDwTxtAddr() {
+					continue
+				}
+				rsym := r.Sym()
+				rst := d.ldr.SymType(rsym)
+				// Do some consistency checks.
+				if !rst.IsText() {
+					// R_DWTXTADDR_* relocation should only refer to
+					// text symbols, so something apparently went
+					// wrong here.
+					log.Fatalf("internal error: R_DWTXTADDR_* relocation on dwinfosym for %s against non-function %s type:%s", d.ldr.SymName(fnSym), d.ldr.SymName(rsym), rst.String())
+				}
+				if runit := d.ldr.SymUnit(rsym); runit != unit {
+					log.Fatalf("internal error: R_DWTXTADDR_* relocation target text sym unit mismatch (want %q got %q)", unit.Lib.Pkg, runit.Lib.Pkg)
+				}
+				d.assignDebugAddrSlot(unit, fnSym, r, dasu)
+			}
+		}
+	}
 }
 
 func (d *dwctxt) dwarfGenerateDebugSyms() {
 	abbrevSec := d.writeabbrev()
 	dwarfp = append(dwarfp, abbrevSec)
 	d.calcCompUnitRanges()
-	sort.Sort(compilationUnitByStartPC(d.linkctxt.compUnits))
+	slices.SortFunc(d.linkctxt.compUnits, compilationUnitByStartPCCmp)
 
 	// newdie adds DIEs to the *beginning* of the parent's DIE list.
 	// Now that we're done creating DIEs, reverse the trees so DIEs
@@ -1998,6 +2320,7 @@ func (d *dwctxt) dwarfGenerateDebugSyms() {
 		s.SetReachable(true)
 		return s.Sym()
 	}
+
 	mkAnonSym := func(kind sym.SymKind) loader.Sym {
 		s := d.ldr.MakeSymbolUpdater(d.ldr.CreateExtSym("", 0))
 		s.SetType(kind)
@@ -2007,17 +2330,37 @@ func (d *dwctxt) dwarfGenerateDebugSyms() {
 
 	// Create the section symbols.
 	frameSym := mkSecSym(".debug_frame")
-	locSym := mkSecSym(".debug_loc")
 	lineSym := mkSecSym(".debug_line")
-	rangesSym := mkSecSym(".debug_ranges")
+	var rangesSym, locSym loader.Sym
+	if buildcfg.Experiment.Dwarf5 {
+		rangesSym = mkSecSym(".debug_rnglists")
+		locSym = mkSecSym(".debug_loclists")
+	} else {
+		rangesSym = mkSecSym(".debug_ranges")
+		locSym = mkSecSym(".debug_loc")
+	}
 	infoSym := mkSecSym(".debug_info")
+	var addrSym loader.Sym
+	if buildcfg.Experiment.Dwarf5 {
+		addrSym = mkSecSym(".debug_addr")
+	}
 
 	// Create the section objects
 	lineSec := dwarfSecInfo{syms: []loader.Sym{lineSym}}
-	locSec := dwarfSecInfo{syms: []loader.Sym{locSym}}
-	rangesSec := dwarfSecInfo{syms: []loader.Sym{rangesSym}}
 	frameSec := dwarfSecInfo{syms: []loader.Sym{frameSym}}
 	infoSec := dwarfSecInfo{syms: []loader.Sym{infoSym}}
+	var addrSec, rangesSec, locSec dwarfSecInfo
+	if buildcfg.Experiment.Dwarf5 {
+		addrHdr := d.writeDebugAddrHdr()
+		addrSec.syms = []loader.Sym{addrSym, addrHdr}
+		rnglistsHdr := d.writeDebugRngListsHdr()
+		rangesSec.syms = []loader.Sym{rangesSym, rnglistsHdr}
+		loclistsHdr := d.writeDebugLocListsHdr()
+		locSec.syms = []loader.Sym{locSym, loclistsHdr}
+	} else {
+		rangesSec = dwarfSecInfo{syms: []loader.Sym{rangesSym}}
+		locSec = dwarfSecInfo{syms: []loader.Sym{locSym}}
+	}
 
 	// Create any new symbols that will be needed during the
 	// parallel portion below.
@@ -2028,6 +2371,7 @@ func (d *dwctxt) dwarfGenerateDebugSyms() {
 		us.lineProlog = mkAnonSym(sym.SDWARFLINES)
 		us.rangeProlog = mkAnonSym(sym.SDWARFRANGE)
 		us.infoEpilog = mkAnonSym(sym.SDWARFFCN)
+		us.addrsym = mkAnonSym(sym.SDWARFADDR)
 	}
 
 	var wg sync.WaitGroup
@@ -2069,6 +2413,40 @@ func (d *dwctxt) dwarfGenerateDebugSyms() {
 		return syms
 	}
 
+	patchHdr := func(sec *dwarfSecInfo, len uint64) {
+		hdrsym := sec.syms[1]
+		len += uint64(d.ldr.SymSize(hdrsym))
+		su := d.ldr.MakeSymbolUpdater(hdrsym)
+		if isDwarf64(d.linkctxt) {
+			len -= 12                          // sub size of length field
+			su.SetUint(d.arch, 4, uint64(len)) // 4 because of 0XFFFFFFFF
+		} else {
+			len -= 4 // subtract size of length field
+			su.SetUint32(d.arch, 0, uint32(len))
+		}
+	}
+
+	if buildcfg.Experiment.Dwarf5 {
+		// Compute total size of the DWARF5-specific .debug_* syms in
+		// each compilation unit.
+		var rltot, addrtot, loctot uint64
+		for i := 0; i < ncu; i++ {
+			addrtot += uint64(d.ldr.SymSize(unitSyms[i].addrsym))
+			rs := unitSyms[i].rangessyms
+			for _, s := range rs {
+				rltot += uint64(d.ldr.SymSize(s))
+			}
+			loc := unitSyms[i].locsyms
+			for _, s := range loc {
+				loctot += uint64(d.ldr.SymSize(s))
+			}
+		}
+		// Call a helper to patch the length field in the headers.
+		patchHdr(&addrSec, addrtot)
+		patchHdr(&rangesSec, rltot)
+		patchHdr(&locSec, loctot)
+	}
+
 	// Stitch together the results.
 	for i := 0; i < ncu; i++ {
 		r := &unitSyms[i]
@@ -2076,6 +2454,9 @@ func (d *dwctxt) dwarfGenerateDebugSyms() {
 		infoSec.syms = append(infoSec.syms, markReachable(r.infosyms)...)
 		locSec.syms = append(locSec.syms, markReachable(r.locsyms)...)
 		rangesSec.syms = append(rangesSec.syms, markReachable(r.rangessyms)...)
+		if buildcfg.Experiment.Dwarf5 && r.addrsym != 0 {
+			addrSec.syms = append(addrSec.syms, r.addrsym)
+		}
 	}
 	dwarfp = append(dwarfp, lineSec)
 	dwarfp = append(dwarfp, frameSec)
@@ -2088,6 +2469,9 @@ func (d *dwctxt) dwarfGenerateDebugSyms() {
 		dwarfp = append(dwarfp, locSec)
 	}
 	dwarfp = append(dwarfp, rangesSec)
+	if buildcfg.Experiment.Dwarf5 {
+		dwarfp = append(dwarfp, addrSec)
+	}
 
 	// Check to make sure we haven't listed any symbols more than once
 	// in the info section. This used to be done by setting and
@@ -2096,7 +2480,8 @@ func (d *dwctxt) dwarfGenerateDebugSyms() {
 	seen := loader.MakeBitmap(d.ldr.NSym())
 	for _, s := range infoSec.syms {
 		if seen.Has(s) {
-			log.Fatalf("symbol %s listed multiple times", d.ldr.SymName(s))
+			log.Fatalf("dwarf symbol %s listed multiple times",
+				d.ldr.SymName(s))
 		}
 		seen.Set(s)
 	}
@@ -2129,7 +2514,13 @@ func dwarfaddshstrings(ctxt *Link, add func(string)) {
 		return
 	}
 
-	secs := []string{"abbrev", "frame", "info", "loc", "line", "gdb_scripts", "ranges"}
+	secs := []string{"abbrev", "frame", "info", "loc", "line", "gdb_scripts"}
+	if buildcfg.Experiment.Dwarf5 {
+		secs = append(secs, "addr", "rnglists", "loclists")
+	} else {
+		secs = append(secs, "ranges", "loc")
+	}
+
 	for _, sec := range secs {
 		add(".debug_" + sec)
 		if ctxt.IsExternal() {
@@ -2247,21 +2638,16 @@ func dwarfcompress(ctxt *Link) {
 	Segdwarf.Length = pos - Segdwarf.Vaddr
 }
 
-type compilationUnitByStartPC []*sym.CompilationUnit
-
-func (v compilationUnitByStartPC) Len() int      { return len(v) }
-func (v compilationUnitByStartPC) Swap(i, j int) { v[i], v[j] = v[j], v[i] }
-
-func (v compilationUnitByStartPC) Less(i, j int) bool {
+func compilationUnitByStartPCCmp(a, b *sym.CompilationUnit) int {
 	switch {
-	case len(v[i].Textp) == 0 && len(v[j].Textp) == 0:
-		return v[i].Lib.Pkg < v[j].Lib.Pkg
-	case len(v[i].Textp) != 0 && len(v[j].Textp) == 0:
-		return true
-	case len(v[i].Textp) == 0 && len(v[j].Textp) != 0:
-		return false
+	case len(a.Textp) == 0 && len(b.Textp) == 0:
+		return strings.Compare(a.Lib.Pkg, b.Lib.Pkg)
+	case len(a.Textp) != 0 && len(b.Textp) == 0:
+		return -1
+	case len(a.Textp) == 0 && len(b.Textp) != 0:
+		return +1
 	default:
-		return v[i].PCs[0].Start < v[j].PCs[0].Start
+		return cmp.Compare(a.PCs[0].Start, b.PCs[0].Start)
 	}
 }
 
@@ -2291,3 +2677,34 @@ func addDwsectCUSize(sname string, pkgname string, size uint64) {
 	defer dwsectCUSizeMu.Unlock()
 	dwsectCUSize[sname+"."+pkgname] += size
 }
+
+// writeDebugMiscSecHdr writes a header section for the new new DWARF5
+// sections ".debug_addr", ".debug_loclists", and ".debug_rnglists".
+// A description of the format/layout of these headers can be found in
+// the DWARF5 spec in sections 7.27 (.debug_addr), 7.28
+// (.debug_rnglists) and 7.29 (.debug_loclists).
+func (d *dwctxt) writeDebugMiscSecHdr(st sym.SymKind, addOffsetEntryCount bool) loader.Sym {
+	su := d.ldr.MakeSymbolUpdater(d.ldr.CreateExtSym("", 0))
+	su.SetType(st)
+	su.SetReachable(true)
+	d.createUnitLength(su, 0)          // will be filled in later.
+	su.AddUint16(d.arch, 5)            // dwarf version (appendix F)
+	su.AddUint8(uint8(d.arch.PtrSize)) // address_size
+	su.AddUint8(0)                     // segment selector
+	if addOffsetEntryCount {
+		su.AddUint32(d.arch, 0) // offset entry count (required but unused)
+	}
+	return su.Sym()
+}
+
+func (d *dwctxt) writeDebugRngListsHdr() loader.Sym {
+	return d.writeDebugMiscSecHdr(sym.SDWARFRANGE, true)
+}
+
+func (d *dwctxt) writeDebugLocListsHdr() loader.Sym {
+	return d.writeDebugMiscSecHdr(sym.SDWARFLOC, true)
+}
+
+func (d *dwctxt) writeDebugAddrHdr() loader.Sym {
+	return d.writeDebugMiscSecHdr(sym.SDWARFADDR, false)
+}
